/* 	I2C routine library for the PIC Micro

	Version 1.0
	Started: 21/4/00
	Last Edit: 21/4/00
	Copyright(c)2000 David L. Jones

NOTES:
- Direct PORT code, does not use the internal I2C module on those PIC's that 
	have it
- No bus arbitration ability, simple PIC=Master/Multiple Slave interface only
- No multi-master capability
- No timing ability, so I2C speed will depend on PIC clock rate
- The pull-up resistors on the IC2 pins are not enabled, you must do 
	it yourself if you want to avoid using external resistors
- make sure the I2C_init routine is run at least once before anything else
- These routines default to using RA0=SCL and RA1=SDA, change the define 
	statements below and the I2C_init routine to use other pins
- code is generic and can be compiled to any target PIC device without
	modification

ROUTINES:
I2C_start() - Sets the I2C bus to the START condition
I2C_stop() - Sets the I2C bus to the STOP condition
I2C_stable() - Sets the I2C bus to the STABLE (IDLE) condition
char I2C_readbyte() - returns a byte from the I2C bus
I2C_sendbyte() - sends a byte down the I2C bus
I2C_init() - Initialises stuff

HOW TO TALK TO AN I2C DEVICE:
(generically, please refer to the chip datasheet as not all I2C chips work 
this way)
(remember to run I2C_init first!)

To set a byte within the register of an I2C device:
	I2C_stable
	I2C_start
	I2C_sendbyte(device address with WRITE bit set)
	I2C_sendbyte(register address)
	I2C_sendbyte(byte to write to register)
	...write subsequent sequential bytes (if any)
	I2C_stop

To read the current register value from an I2C device:
	I2C_stable
	I2C_start
	I2C_sendbyte(device address with READ bit set)
	byte=I2C_readbyte	
	I2C_stop

To change the current register pointer in an I2C device:
	I2C_stable
	I2C_start
	I2C_sendbyte(device address with WRITE bit set)
	I2C_sendbyte(new pointer address)
	I2C_stop
*/

/*********************************************************
This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
**********************************************************/

#define set_scl LATB=LATB|OR8		/*SCL pin HIGH definition. Change to your own pin! */
#define clr_scl LATB=LATB&AND8		/*SCL pin LOW definition. Change to your own pin! */
#define set_sda _TRISB9=1			/*SDA pin HIGH definition. 
	Change to your own pin!
	pin must be set to INPUT via TRIS because I2C data line is open 
	collector */
#define clr_sda	_TRISB9=0			/*SDA pin HIGH definition.
	This macro only sets the SDA pin to an output, you still have to 
	manually ensure that the SDA pin PORT register is set LOW 
	This is doen by running the init_I2C routine */
#define sda_data _RB9		/*the SDA input data. Change to your own pin! */
#define high=1				/*HIGH bus state definition */
#define low=0				/*LOW bus state definition */

/* *******************************************************************
	This routine must be run at least once in your code before
	any of the I2C routines below. No need to run it more than once
	unless your other PIC code is changing your pins! */
void init_I2C() {
	LATB=LATB&AND9;		/* set the SDA pin LOW. The SDA pin is then set 
			HIGH by the TRIS command */
	_TRISB8=0;	/* set SCL line to an OUPUT */
	_TRISB9=1;	/* set SDA line to an INPUT just to be sure! */
}
	
/* *******************************************************************
	The "start" routine sets the I2C bus to the START condition */
void I2C_start() {
	set_scl;
	set_sda;
	clr_sda;
	clr_scl;		
}

/* *******************************************************************
	The "stop" routine sets the I2C bus to the STOP condition */
void I2C_stop() {
	clr_sda;
	set_scl;
	set_sda;
	set_scl;		
}

/* *******************************************************************
	The "stable" routine sets the I2C bus to the STABLE condition */
void I2C_stable() {
	set_sda;
	set_scl;
}

/* *******************************************************************
	"sendbyte" sends a byte down the I2C bus
	The byte to be sent is passed to the routine via "sendbuf" */
void I2C_sendbyte(unsigned char sendbuf) {
	clr_scl;			/* set scl low initially */

	/* note:I2C requires MSBit to be sent first */
	
	if ((sendbuf&128)==0) clr_sda; else set_sda;	/* ouput bit7 */
	set_scl;				/* toggle clock line */
	clr_scl;

	if ((sendbuf&64)==0) clr_sda; else set_sda;	/* ouput bit6 */
	set_scl;				/* toggle clock line */
	clr_scl;

	if ((sendbuf&32)==0) clr_sda; else set_sda;	/* ouput bit5 */
	set_scl;				/* toggle clock line */
	clr_scl;

	if ((sendbuf&16)==0) clr_sda; else set_sda;	/* ouput bit4 */
	set_scl;				/* toggle clock line */
	clr_scl;

	if ((sendbuf&8)==0) clr_sda; else set_sda;	/* ouput bit3 */
	set_scl;				/* toggle clock line */
	clr_scl;

	if ((sendbuf&4)==0) clr_sda; else set_sda;	/* ouput bit2 */
	set_scl;				/* toggle clock line */
	clr_scl;

	if ((sendbuf&2)==0) clr_sda; else set_sda;	/* ouput bit1 */
	set_scl;				/* toggle clock line */
	clr_scl;

	if ((sendbuf&1)==0) clr_sda; else set_sda;	/* ouput bit0 */
	set_scl;				/* toggle clock line */
	clr_scl;
	
	set_sda;
	set_scl;	/* allow SLAVE to return ACK but just ignore it */
	clr_scl;
}

/* *******************************************************************
	"readbyte" reads a byte from the I2C bus 
	Function returns the byte received
	NOTE: This routine has no timeout, so may get stuck in an infinite
	loop if it does not get the correct data from the I2C port */
unsigned char I2C_readbyte() {
	unsigned char byteread;
	byteread=0;	/* assume the data will be 0 */
	clr_scl;
	set_sda;	/*set SDA line to INPUT */

	set_scl;
	if (sda_data==1) byteread+=128;		/* get bit7 */
	clr_scl;

	set_scl;
	if (sda_data==1) byteread+=64;		/* get bit6 */
	clr_scl;

	set_scl;
	if (sda_data==1) byteread+=32;		/* get bit5 */
	clr_scl;

	set_scl;
	if (sda_data==1) byteread+=16;		/* get bit4 */
	clr_scl;

	set_scl;
	if (sda_data==1) byteread+=8;		/* get bit3 */
	clr_scl;

	set_scl;
	if (sda_data==1) byteread+=4;		/* get bit2 */
	clr_scl;

	set_scl;
	if (sda_data==1) byteread+=2;		/* get bit1 */
	clr_scl;

	set_scl;
	if (sda_data==1) byteread+=1;		/* get bit0 */
	clr_scl;
		
	return byteread;	/* return the byte read from the bus */
}

